<?php

namespace Shavshukov\EventAppBase\Http\Routing;

use Shavshukov\EventAppBase\FileReaders\FileReaderInterface;

use Psr\Http\Message\ServerRequestInterface;
use React\Http\Response;
use React\Promise;

class Matcher
{
    /**
     * File reader
     *
     * @var \Shavshukov\EventAppBase\FileReaders\FileReaderInterface
     */
    protected $fileReader;
    
    /**
     *
     * @var \Shavshukov\EventAppBase\Http\Routing\Router
     */
    protected $router;
    
    /**
     *
     * @var \Psr\Http\Message\ServerRequestInterface 
     */
    protected $request;
    
    /**
     *
     * @var string
     */
    protected $method;
    
    /**
     *
     * @var string
     */
    protected $path;
    
    /**
     * Constructor
     *
     * @param \Shavshukov\EventAppBase\FileReaders\FileReaderInterface $fileReader
     * @param \Shavshukov\EventAppBase\Http\Routing\Router $router
     * @param \Psr\Http\Message\ServerRequestInterface $request
     */
    public function __construct(FileReaderInterface $fileReader, Router $router, ServerRequestInterface $request)
    {
        $this->fileReader = $fileReader;
        $this->router = $router;
        $this->request = $request;
        $this->method = $request->getMethod();
        $this->path = $this->getRoutePath($request);
    }
    
    /**
     * Match route to the action or a response until either a route or response is found, or a rejection is raised
     * 
     * @return \React\Promise\PromiseInterface<array|\React\Http\Response>
     */
    public function match(): Promise\PromiseInterface
    {
        $matchMore = function ($result = null) use (&$matchMore) {
            if ($result) {
                return $result;
            }
            
            return $this->tryMatch()->then($matchMore);
        };
        
        return $this->tryMatch()->then($matchMore);
    }
    
    /**
     * Attempt to match route to the action or a response
     * 
     * @return \React\Promise\PromiseInterface<null|array|\React\Http\Response>
     */
    protected function tryMatch(): Promise\PromiseInterface
    {
        // a string -> redirection; a null -> no route, an array -> route exists
        $route = $this->router->getRoute($this->path, $this->method);

        if (is_string($route) && $this->path !== $route) {
            $this->path = $route;
            return $this->authorisationCheck();
        }
        
        // route is not found for path, however the path could be a downloadable file
        if ($route === null) {
            // even without a found route, a path must be checked for authorisation, so 403 has more priority than 404
            return $this->authorisationCheck(function () {
                // in case of GET, check for file download, and if no success, set not found route
                // for other methods just set it immediately
                if ($this->path && $this->path !== '/' && $this->method === 'GET') {
                    return $this->checkForFileDownload();
                } else {
                    return $this->setNotFoundRoute();
                }
            });
        }
        
        // if the $route is string, the cycle only breaks on equal path (already same route) -> recursion
        if (is_string($route)) {
            return Promise\reject(new \Exception("Recursive routing detected: {$this->method} {$this->path} is redirected to itself"));
        }

        // bad routes
        if (!is_array($route) || count($route) < 2) {
            return Promise\reject(new \Exception("A route must be either a string (alias for other route),"
                . " or an array ['ControllerClass', 'actionMethod'], wrong route at path: {$this->path}, method: {$this->method}"));
        }
        
        // now we have an actual valid route, lets make the check return it as a result
        return $this->authorisationCheck(function () use ($route) {
            return $route;
        });
    }
    
    /**
     * Get paths and attempt to resolve them to file contents; send a response with it, should success
     * If user path to the file from the public folder does not match server path to the file from the public folder, the access is denied
     * 
     * @return \React\Promise\PromiseInterface<null|\React\Http\Response>
     */
    protected function checkForFileDownload(): Promise\PromiseInterface
    {
        $publicRealPath = $this->router->getPublicRealPath();
        $downloadPath = $publicRealPath . str_replace('/', DIRECTORY_SEPARATOR, $this->path);
        $downloadRealPath = realpath($downloadPath);
        
        // false means the file does not exists
        if ($downloadRealPath === false) {
            return $this->setNotFoundRoute();
        }

        // deny any access by path injection
        if ($downloadPath !== $downloadRealPath) {
            return $this->setAccessDeniedRoute();
        }

        return $this->downloadFile($downloadPath)->then(function ($d) {
            return $d;
        });
    }
    
    /**
     * Attempt to download file, or use not found way if it doesn't exists
     * 
     * @param string $downloadPath
     * @return \React\Promise\PromiseInterface<\React\Http\Response>
     */
    protected function downloadFile(string $downloadPath): Promise\PromiseInterface
    {
        return Promise\resolve(new Response(
            200,
            ['Content-Type' => mime_content_type($downloadPath)],
            // find any better way..?(
            $this->fileReader->asResponseArgument($downloadPath)
        ));
    }
    
    /**
     * Change method and path to serve 403 error page
     * 
     * @return \React\Promise\PromiseInterface
     */
    protected function setAccessDeniedRoute(): Promise\PromiseInterface
    {
        $this->path = '/403';
        $this->method = 'GET';
        return $this->authorisationCheck();
    }
    
    /**
     * Change method and path to serve 404 error page
     * 
     * @return \React\Promise\PromiseInterface
     */
    protected function setNotFoundRoute(): Promise\PromiseInterface
    {
        $this->path = '/404';
        $this->method = 'GET';
        return $this->authorisationCheck();
    }
    
    /**
     * Check authorisation through router by the request, method and path
     * Should the authorisation succeed, call the callback and return a result of the callback as the promise value
     * Otherwise, change path and method and return nothing as the promise value
     * 
     * @param \Closure $successCallback
     * @return \React\Promise\PromiseInterface
     */
    protected function authorisationCheck($successCallback = null): Promise\PromiseInterface
    {
        return $this->router->authorise($this->request, $this->method, $this->path)->then($successCallback, function ($redirection = null) {
            list($method, $path) = ($redirection ?? ['GET', '/403']);

            // detect another recursion: if the redirection method and path are equal to current, then authentication redirection is wrong
            if ($this->path === $path && $this->method === $method) {
                return Promise\reject(new \Exception("Recursive routing detected: {$this->method} {$this->path}"
                    . " is requested as a redirection from unsuccessful authentication to itself"));
            }

            $this->path = $path;
            $this->method = $method;
        });
    }
    
    /**
     * Get route path string from request
     * 
     * @param \Psr\Http\Message\ServerRequestInterface $request
     * @return string
     */
    protected function getRoutePath(ServerRequestInterface $request): string
    {
        return rtrim(rtrim($request->getUri()->getPath()), '/') ?: '/';
    }
    
}
