<?php

namespace Shavshukov\EventAppBase\Model;

use Shavshukov\EventAppBase\Storage\Storage;
use Shavshukov\EventAppBase\Interfaces\StorageInterface;
use Shavshukov\EventAppBase\Interfaces\ModelInterface;

abstract class StorableModel extends Storage implements ModelInterface
{
    use ModelDataTrait;
    use ModelValidationTrait;
    
    /**
     * Array data
     *
     * @var array
     */
    protected $data = [];
    
    /**
     * Extend model by storage or another model
     * 
     * @param \Shavshukov\EventAppBase\Interfaces\StorageInterface $storage
     * @return \static
     */
    public static function extend(StorageInterface $storage)
    {
        $newModel = new static();
        $newModel->setBackEnd($storage);
        $newModel->source = $storage->content();
        $newModel->setDataAfterSourceChanged(true);
        return $newModel;
    }
    
    /**
     * {@inheritDoc}
     */
    public function content()
    {
        return $this->data;
    }
    
    /**
     * {@inheritDoc}
     */
    public function load(): bool
    {
        return $this->setDataAfterSourceChanged(parent::load());
    }
    
    /**
     * {@inheritDoc}
     */
    public function modifyBy(StorageInterface $storage): bool
    {
        $data = $storage->source();
        $this->data = $data;
        $this->modified = true;
        return $this->save();
    }
    
    /**
     * {@inheritDoc}
     */
    public function loadAsync(): \React\Promise\PromiseInterface
    {
        return parent::loadAsync()->then(function($result) {
            return $this->setDataAfterSourceChanged($result);
        });
    }
    
    /**
     * {@inheritDoc}
     */
    public function save(): bool
    {
        if ($this->modified) {
            $this->source = $this->encode();
            return $this->saveBackEnd();
        }
        
        return false;
    }
    
    /**
     * Set new data after the source is changed
     * 
     * @param bool $result
     */
    protected function setDataAfterSourceChanged($result)
    {
        $this->data = $result ? $this->decode() : [];
        $this->modified = false;
        return $result;
    }
    
    /**
     * Decode source into an array (if necessary, may return the same array)
     * 
     * @return array
     */
    abstract protected function decode(): array;
    
    /**
     * Encode source back into the desired type of the storage
     * 
     * @return mixed
     */
    abstract protected function encode();
    
}
